(*----------------------------------------------------------------------------*
 *                                                                            *
 *  This unit includes TD3DNitroFont class for rendering text using           *
 *  bitmap based font in Direct3D 8.x. Used font format is taken from Nitro   *
 *  [http://www.sulaco.co.za/nitrogen/],where you can find Font Studio - for  *
 *  generating font files and OpenGL font renderer demo                       *
 *                                                                            *
 *  Realization by Alexey Barkovoy                                            *
 *  E-Mail: clootie@reactor.ru                                                *
 *                                                                            *
 *  Modified: 17-Mar-2002                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru                                                *
 *----------------------------------------------------------------------------*)

unit NitroFnt;

interface

{$MINENUMSIZE 4}
{$ALIGN ON}

uses
  Windows, Direct3D8;

type
  // Font rendering flags
  TFontRenderFlag = (rfCentered, rfTwoSided, rfFiltered);
  TFontRenderFlags = set of TFontRenderFlag;

  TCharDef = packed record
    T1,U1,T2,U2, W, H: Single;
  end;

  PCharDefArray = ^TCharDefArray;
  TCharDefArray = array [0..(MaxInt div SizeOf(TCharDef))-1] of TCharDef;

  TD3DNitroFont = class
  private
    m_strFontBitmapFile: array[0..79] of Char; // Name of bitmap image with font
    fZEnable: Boolean;           // Do render to Z-buffer

    fWidth, fHeight: DWORD;        // Sizes of Font Texture - used in 2D rendering

    // Font Instance data
    StartChar, FontLen, CharOffset: Integer;
    SpaceWidth: Single;
    FontChars: PCharDefArray;

    // DirectX interfaces
    m_pd3dDevice: IDirect3DDevice8;       // A D3DDevice used for rendering
    m_pTexture: IDirect3DTexture8;        // The d3d texture for this font
    m_pVB: IDirect3DVertexBuffer8;        // VertexBuffer for rendering text

    // Stateblocks for setting and restoring render states
    m_dwSavedStateBlock: DWORD;
    m_dwDrawTextStateBlock: DWORD;
    procedure GenerateCharCoords(var c: Char; out tx1, ty1, tx2, ty2: Single);
  public
    // State management helpers
    function PrepareDrawing(Flags: TFontRenderFlags): HRESULT;
    procedure RestoreState;

    function TextLen(const Txt: String): Single;
    // 2D and 3D text drawing functions
    function DrawText(x, y: Single; dwColor: DWORD;
                      strText: PChar; Flags: TFontRenderFlags): HRESULT;
    function DrawTextScaled(x, y, z: Single; fXScale, fYScale: Single; dwColor: DWORD;
                            strText: PChar; Flags: TFontRenderFlags): HRESULT;
    function Render3DText(strText: PChar; Flags: TFontRenderFlags): HRESULT;
    procedure Draw(const X,Y: single; const Txt: String; Lev: Single;
      Flags: TFontRenderFlags;
      DoNotManageStates: Boolean);

    // Initializing and destroying device-dependent objects
    function InitDeviceObjects(pd3dDevice: IDirect3DDevice8): HRESULT;
    function RestoreDeviceObjects: HRESULT;
    function InvalidateDeviceObjects: HRESULT;
    function DeleteDeviceObjects: HRESULT;

    // Constructor / destructor
    constructor Create(strFontBitmapFile: PChar);
    destructor Destroy; override;

    property ZEnable: Boolean read fZEnable write fZEnable default True;
    property Width: DWORD read fWidth;
    property Height: DWORD read fHeight;
  end;

implementation

uses
  SysUtils,
  D3DX8;

//-----------------------------------------------------------------------------
// Custom vertex types for rendering text
//-----------------------------------------------------------------------------
const
  MAX_NUM_VERTICES = 50*6;

type
  PFont2DVertex = ^TFont2DVertex;
  TFont2DVertex = packed record
    p: TD3DXVector4;
    color: DWORD;
    tu, tv: Single;
  end;

  PFont3DVertex = ^TFont3DVertex;
  TFont3DVertex = packed record
    p: TD3DXVector3;
    n: TD3DXVector3;
    case Boolean of
    true:
      (tu, tv: Single;);
    false:
      (t1: TD3DXVector2;);
  end;

const
  D3DFVF_Font2DVertex = (D3DFVF_XYZRHW or D3DFVF_DIFFUSE or D3DFVF_TEX1);
  D3DFVF_Font3DVertex = (D3DFVF_XYZ or D3DFVF_NORMAL or D3DFVF_TEX1);

function InitFont2DVertex(const p: TD3DXVector4; color: TD3DColor;
  tu, tv: Single): TFont2DVertex;
begin
  Result.p := p;
  Result.color := color;
  Result.tu := tu;
  Result.tv := tv;
end;

function InitFont3DVertex(const p: TD3DXVector3; const n: TD3DXVector3;
  tu, tv: Single): TFont3DVertex;
begin
  Result.p := p;
  Result.n := n;
  Result.tu := tu;
  Result.tv := tv;
end;


{ TD3DTextureFont }

//-----------------------------------------------------------------------------
// Name: TD3DTextureFont.Create
// Desc: Font class constructor creates object with font from strFontBitmapFile
//       using color ColorKey as transparent (background) color
//-----------------------------------------------------------------------------
constructor TD3DNitroFont.Create(strFontBitmapFile: PChar);
begin
  StrCopy(m_strFontBitmapFile, strFontBitmapFile);
  fZEnable:= True;

  m_pd3dDevice          := nil;
  m_pTexture            := nil;
  m_pVB                 := nil;

  m_dwSavedStateBlock    := 0;
  m_dwDrawTextStateBlock := 0;

end;

destructor TD3DNitroFont.Destroy;
begin
  InvalidateDeviceObjects;
  DeleteDeviceObjects;
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initializes device-dependent objects, including the vertex buffer used
//       for rendering text and the texture map which stores the font image.
//-----------------------------------------------------------------------------
function TD3DNitroFont.InitDeviceObjects(
  pd3dDevice: IDirect3DDevice8): HRESULT;
var
  SrcInfo: TD3DXImageInfo;
  fHandle, fMemHandle: THandle;
  pMemFile, p: PDWORD;
  dFileSize: Integer;
  Stpos: DWORD;
  i: Integer;
  FileFontChars: PCharDefArray;
begin
  // Keep a local copy of the device
  m_pd3dDevice := pd3dDevice;

  fHandle:= CreateFile(m_strFontBitmapFile, GENERIC_READ, FILE_SHARE_READ, nil,
    OPEN_EXISTING, 0, 0);
  if (fHandle = INVALID_HANDLE_VALUE) then
  begin
    Result:= E_FAIL;
    Exit;
  end;

  fMemHandle:= CreateFileMapping(fHandle, nil, PAGE_READONLY, 0, 0, nil);
  if (fMemHandle = 0) then
  begin
    Result:= E_FAIL;
    Exit;
  end;

  pMemFile:= MapViewOfFile(fMemHandle, FILE_MAP_READ, 0, 0, 0);
  if (pMemFile = nil) then
  begin
    Result:= E_FAIL;
    Exit;
  end;

  dFileSize:= GetFileSize(fHandle, nil);

  // Load texture
  Result:= D3DXCreateTextureFromFileInMemoryEx(m_pd3dDevice, pMemFile^, dFileSize,
    D3DX_DEFAULT, D3DX_DEFAULT, D3DX_DEFAULT, 0, D3DFMT_UNKNOWN, D3DPOOL_MANAGED,
    D3DX_DEFAULT, D3DX_DEFAULT, 0, @SrcInfo, nil, m_pTexture);
  if Failed(Result) then Exit;

  // Keep texture dimentions for future 2D font rendering
  fWidth:= SrcInfo.Width;
  fHeight:= SrcInfo.Height;

  p:= PDWORD(Integer(pMemFile) + dFileSize - SizeOf(DWORD));

  Stpos:= P^;

  p:= PDWORD(DWORD(pMemFile) + Stpos);

  FontLen := p^;        Inc(p);
  StartChar := p^;      Inc(p);
  CharOffset := p^;     Inc(p);

  GetMem(FontChars, SizeOf(TCharDef)*(FontLen+1));
  FileFontChars:= PCharDefArray(p);

  SpaceWidth:= 0;
  for i := 0 to FontLen do
  begin
    FontChars^[i]:= FileFontChars^[i];
    SpaceWidth:= SpaceWidth + FontChars^[i].T2 - FontChars^[i].T1;
  end;
  SpaceWidth:= SpaceWidth / FontLen;

  UnmapViewOfFile(pMemFile);
  CloseHandle(fHandle);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
function TD3DNitroFont.RestoreDeviceObjects: HRESULT;
var
  which: LongWord;
begin
  // Create vertex buffer for the letters
  Result:= m_pd3dDevice.CreateVertexBuffer(MAX_NUM_VERTICES*SizeOf(TFont2DVertex),
                                           D3DUSAGE_WRITEONLY or D3DUSAGE_DYNAMIC,
                                           0, D3DPOOL_DEFAULT, m_pVB);
  if FAILED(Result) then Exit;

  // Create the state blocks for rendering text
  for which:= 0 to 1 do
  begin
    m_pd3dDevice.BeginStateBlock;
    m_pd3dDevice.SetTexture(0, m_pTexture);

    if ZEnable then
      m_pd3dDevice.SetRenderState(D3DRS_ZENABLE, iTRUE)
    else
      m_pd3dDevice.SetRenderState(D3DRS_ZENABLE, iFALSE);

    m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iTrue);
    m_pd3dDevice.SetRenderState(D3DRS_SRCBLEND,         D3DBLEND_SRCALPHA);
    m_pd3dDevice.SetRenderState(D3DRS_DESTBLEND,        D3DBLEND_INVSRCALPHA);
    m_pd3dDevice.SetRenderState(D3DRS_ALPHATESTENABLE,  iTrue);
    m_pd3dDevice.SetRenderState(D3DRS_ALPHAREF,         $08);
    m_pd3dDevice.SetRenderState(D3DRS_ALPHAFUNC,        D3DCMP_GREATEREQUAL);
    m_pd3dDevice.SetRenderState(D3DRS_FILLMODE,         D3DFILL_SOLID);
    m_pd3dDevice.SetRenderState(D3DRS_CULLMODE,         D3DCULL_CCW);
    m_pd3dDevice.SetRenderState(D3DRS_STENCILENABLE,    iFalse);
    m_pd3dDevice.SetRenderState(D3DRS_CLIPPING,         iTrue);
    m_pd3dDevice.SetRenderState(D3DRS_EDGEANTIALIAS,    iFalse);
    m_pd3dDevice.SetRenderState(D3DRS_CLIPPLANEENABLE,  iFalse);
    m_pd3dDevice.SetRenderState(D3DRS_VERTEXBLEND,      iFalse);
    m_pd3dDevice.SetRenderState(D3DRS_INDEXEDVERTEXBLENDENABLE, iFalse);
    m_pd3dDevice.SetRenderState(D3DRS_FOGENABLE,        iFalse);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_MODULATE);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAOP,   D3DTOP_MODULATE);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_POINT);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_POINT);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_MIPFILTER, D3DTEXF_NONE);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 0);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_DISABLE);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLOROP,   D3DTOP_DISABLE);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_ALPHAOP,   D3DTOP_DISABLE);

    if (which = 0) then
      m_pd3dDevice.EndStateBlock(m_dwSavedStateBlock)
    else
      m_pd3dDevice.EndStateBlock(m_dwDrawTextStateBlock);
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Destroys all device-dependent objects
//-----------------------------------------------------------------------------
function TD3DNitroFont.InvalidateDeviceObjects: HRESULT;
begin
  m_pVB:= nil;

  // Delete the state blocks
  if (m_pd3dDevice <> nil) then
  begin
    if (m_dwSavedStateBlock <> 0) then
      m_pd3dDevice.DeleteStateBlock(m_dwSavedStateBlock);
    if (m_dwDrawTextStateBlock <> 0) then
      m_pd3dDevice.DeleteStateBlock(m_dwDrawTextStateBlock);
  end;

  m_dwSavedStateBlock    := 0;
  m_dwDrawTextStateBlock := 0;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Destroys all device-dependent objects
//-----------------------------------------------------------------------------
function TD3DNitroFont.DeleteDeviceObjects: HRESULT;
begin
  m_pTexture:= nil;
  m_pd3dDevice:= nil;
  Result:= S_OK;
end;

procedure TD3DNitroFont.GenerateCharCoords(var c: Char; out tx1, ty1, tx2, ty2: Single);
begin
  if (Ord(C) < StartChar) or (Ord(C) >= FontLen + StartChar) then
  begin // generate FAKE symbol
    C:= ' ';
    tx1:= 0; tx2:= SpaceWidth;
    ty1:= 0; ty2:= 0.01/255;
    Exit;
  end;

  tx1:= FontChars^[Ord(C) - StartChar].T1;
  ty1:= FontChars^[Ord(C) - StartChar].U1;
  tx2:= FontChars^[Ord(C) - StartChar].T2 + 0.005;
  ty2:= FontChars^[Ord(C) - StartChar].U2 + 0.005;
end;

function TD3DNitroFont.TextLen(const Txt: String): Single;
var
  I, Let: Integer;
  Len: Single;
begin
  Len := 0;
  For I := 0 to length(Txt) do
  begin
    Let := Ord(Txt[I])-StartChar;
    if (Let <> -1) then
    begin
      Len := Len + FontChars^[Let].W+1-CharOffset;
    end
    else
    Len := Len + SpaceWidth;
  end;
  Result := Len;
end;

function TD3DNitroFont.PrepareDrawing(Flags: TFontRenderFlags): HRESULT;
begin
  if (m_pd3dDevice = nil) then
  begin
    Result:= E_FAIL;
    Exit;
  end;

  // Set up renderstate
  m_pd3dDevice.CaptureStateBlock(m_dwSavedStateBlock);
  m_pd3dDevice.ApplyStateBlock(m_dwDrawTextStateBlock);
  m_pd3dDevice.SetPixelShader(0);

  // Set filter states
  if (rfFiltered in Flags) then
  begin
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
  end;

  Result:= S_OK;
end;

procedure TD3DNitroFont.RestoreState;
begin
  m_pd3dDevice.ApplyStateBlock(m_dwSavedStateBlock);
end;

function TD3DNitroFont.DrawText(x, y: Single; dwColor: DWORD;
  strText: PChar; Flags: TFontRenderFlags): HRESULT;
var
  fStartX: Single;
  pVertices: PFont2DVertex;
  dwNumTriangles: DWORD;
  c: Char;
  tx1, ty1, tx2, ty2: Single;
  w, h: Single;
begin
  Result:= PrepareDrawing(Flags);
  if FAILED(Result) then Exit;

  m_pd3dDevice.SetVertexShader(D3DFVF_FONT2DVERTEX);
  m_pd3dDevice.SetStreamSource(0, m_pVB, SizeOf(TFont2DVertex));

  fStartX := x;

  // Fill vertex buffer
  pVertices:= nil;
  dwNumTriangles := 0;
  m_pVB.Lock(0, 0, PByte(pVertices), D3DLOCK_DISCARD);

  while (strText^ <> #0) do
  begin
    c := strText^;
    Inc(strText);

    if (c = #10) then
    begin
      x := fStartX;
      y := y+(1/8);
    end;
    if (c < ' ') then Continue;

    GenerateCharCoords(c, tx1, ty1, tx2, ty2);

    w := (tx2-tx1)*fWidth;
    h := (ty2-ty1)*fHeight;

    if (c <> ' ') then
    begin
      pVertices^ := InitFont2DVertex(D3DXVector4(x+0-0.5,y+h-0.5,0.9,1.0), dwColor, tx1, ty2); Inc(pVertices);
      pVertices^ := InitFont2DVertex(D3DXVector4(x+0-0.5,y+0-0.5,0.9,1.0), dwColor, tx1, ty1); Inc(pVertices);
      pVertices^ := InitFont2DVertex(D3DXVector4(x+w-0.5,y+h-0.5,0.9,1.0), dwColor, tx2, ty2); Inc(pVertices);
      pVertices^ := InitFont2DVertex(D3DXVector4(x+w-0.5,y+0-0.5,0.9,1.0), dwColor, tx2, ty1); Inc(pVertices);
      pVertices^ := InitFont2DVertex(D3DXVector4(x+w-0.5,y+h-0.5,0.9,1.0), dwColor, tx2, ty2); Inc(pVertices);
      pVertices^ := InitFont2DVertex(D3DXVector4(x+0-0.5,y+0-0.5,0.9,1.0), dwColor, tx1, ty1); Inc(pVertices);
      dwNumTriangles :=dwNumTriangles + 2;

      if dwNumTriangles*3 > (MAX_NUM_VERTICES-6)  then
      begin
        // Unlock, render, and relock the vertex buffer
        m_pVB.Unlock;
        m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLELIST, 0, dwNumTriangles);
        pVertices:= nil;
        m_pVB.Lock(0, 0, PByte(pVertices), D3DLOCK_DISCARD);
        dwNumTriangles := 0;
      end;
    end;

    x := x+w;
  end;

  // Unlock and render the vertex buffer
  m_pVB.Unlock;
  if (dwNumTriangles > 0) then
      m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLELIST, 0, dwNumTriangles);

  // Restore the modified renderstates
  m_pd3dDevice.ApplyStateBlock(m_dwSavedStateBlock);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DrawTextScaled()
// Desc: Draws scaled 2D text.  Note that x and y are in viewport coordinates
//       (ranging from -1 to +1).  fXScale and fYScale are the size fraction
//       relative to the entire viewport.  For example, a fXScale of 0.25 is
//       1/8th of the screen width.  This allows you to output text at a fixed
//       fraction of the viewport, even if the screen or window size changes.
//-----------------------------------------------------------------------------
function TD3DNitroFont.DrawTextScaled(x, y, z, fXScale, fYScale: Single;
  dwColor: DWORD; strText: PChar; Flags: TFontRenderFlags): HRESULT;
var
  vp: TD3DViewport8;
  sx, sy, sz, rhw, fStartX, fLineHeight: Single;
  pVertices: PFont2DVertex;
  dwNumTriangles: DWORD;
  c: Char;
  tx1, ty1, tx2, ty2: Single;
  w, h: Single;
begin
  Result:= PrepareDrawing(Flags);
  if FAILED(Result) then Exit;

  m_pd3dDevice.SetVertexShader(D3DFVF_FONT2DVERTEX);
  m_pd3dDevice.SetStreamSource(0, m_pVB, SizeOf(TFont2DVertex));

  m_pd3dDevice.GetViewport(vp);
  sx  := (x+1.0)*vp.Width/2;
  sy  := (y+1.0)*vp.Height/2;
  sz  := z;
  rhw := 1.0;
  fStartX := sx;

  fLineHeight := (FontChars^[0].U2 - FontChars^[0].U1)*fWidth;

  // Fill vertex buffer

  dwNumTriangles := 0;
  m_pVB.Lock(0, 0, PByte(pVertices), D3DLOCK_DISCARD);

  while (strText^ <> #0) do
  begin
    c := strText^;
    Inc(strText);

    if (c = #10) then
    begin
      sx := fStartX;
      sy := sy + fYScale*vp.Height;
    end;
    if (c < ' ') then Continue;

    GenerateCharCoords(c, tx1, ty1, tx2, ty2);

    w := (tx2-tx1)*fWidth;
    h := (ty2-ty1)*fHeight;

    w := w*(fXScale*vp.Height)/fLineHeight;
    h := h*(fYScale*vp.Height)/fLineHeight;

    if (c <> ' ') then
    begin
      pVertices^ := InitFont2DVertex(D3DXVector4(sx+0-0.5, sy+h-0.5, sz, rhw), dwColor, tx1, ty2); Inc(pVertices);
      pVertices^ := InitFont2DVertex(D3DXVector4(sx+0-0.5, sy+0-0.5, sz, rhw), dwColor, tx1, ty1); Inc(pVertices);
      pVertices^ := InitFont2DVertex(D3DXVector4(sx+w-0.5, sy+h-0.5, sz, rhw), dwColor, tx2, ty2); Inc(pVertices);
      pVertices^ := InitFont2DVertex(D3DXVector4(sx+w-0.5, sy+0-0.5, sz, rhw), dwColor, tx2, ty1); Inc(pVertices);
      pVertices^ := InitFont2DVertex(D3DXVector4(sx+w-0.5, sy+h-0.5, sz, rhw), dwColor, tx2, ty2); Inc(pVertices);
      pVertices^ := InitFont2DVertex(D3DXVector4(sx+0-0.5, sy+0-0.5, sz, rhw), dwColor, tx1, ty1); Inc(pVertices);
      dwNumTriangles := dwNumTriangles + 2;

      if (dwNumTriangles*3 > (MAX_NUM_VERTICES-6)) then
      begin
        // Unlock, render, and relock the vertex buffer
        m_pVB.Unlock;
        m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLELIST, 0, dwNumTriangles);
        m_pVB.Lock(0, 0, PByte(pVertices), D3DLOCK_DISCARD);
        dwNumTriangles := 0;
      end;
    end;

    sx := sx + w;
  end;

  // Unlock and render the vertex buffer
  m_pVB.Unlock;
  if (dwNumTriangles > 0) then
      m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLELIST, 0, dwNumTriangles);

  // Restore the modified renderstates
  m_pd3dDevice.ApplyStateBlock(m_dwSavedStateBlock);

  Result:= S_OK;
end;

function TD3DNitroFont.Render3DText(strText: PChar; Flags: TFontRenderFlags): HRESULT;
var
  x, y: Single;
  fStartX: Single;
  c: Char;
  pVertices: PFont3DVertex;
  dwNumTriangles: DWORD;
  tx1, ty1, tx2, ty2: Single;
  w, h: Single;
begin
  Result:= PrepareDrawing(Flags);
  if FAILED(Result) then Exit;

  // Setup renderstate
  m_pd3dDevice.SetVertexShader(D3DFVF_FONT3DVERTEX);
  m_pd3dDevice.SetStreamSource(0, m_pVB, SizeOf(TFont3DVertex));

  // Position for each text element
  x := 0.0;
  y := 0.0;

  // Center the text block at the origin
  if (rfCentered in Flags) then
  begin
    x := -((DWORD(Length(strText))*fWidth div 16)/10.0)/2.0;
    y := -((fHeight div 8)/10.0)/2.0;
  end;

  // Turn off culling for two-sided text
  if (rfTwoSided in Flags) then
    m_pd3dDevice.SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);

  fStartX := x;

  // Fill vertex buffer
  // dwVertex       := 0;  -  it's not used anyway
  dwNumTriangles := 0;
  m_pVB.Lock(0, 0, PByte(pVertices), D3DLOCK_DISCARD);

  c:= strText^;
  while (c <> #0) do
  begin
    Inc(strText);
    if (c = #10) then
    begin
      x := fStartX;
      y := y-(1/8)*fHeight/10.0;
    end;
    if (Byte(c) < 32) then Continue;

    GenerateCharCoords(c, tx1, ty1, tx2, ty2);

    w := (tx2-tx1) * fWidth  / ( 10.0 );
    h := (ty2-ty1) * fHeight / ( 10.0 );

    if (c <> ' ') then
    begin
      pVertices^ := InitFont3DVertex(D3DXVector3(x+0,y+0,0), D3DXVector3(0,0,-1), tx1, ty2); Inc(pVertices);
      pVertices^ := InitFont3DVertex(D3DXVector3(x+0,y+h,0), D3DXVector3(0,0,-1), tx1, ty1); Inc(pVertices);
      pVertices^ := InitFont3DVertex(D3DXVector3(x+w,y+0,0), D3DXVector3(0,0,-1), tx2, ty2); Inc(pVertices);
      pVertices^ := InitFont3DVertex(D3DXVector3(x+w,y+h,0), D3DXVector3(0,0,-1), tx2, ty1); Inc(pVertices);
      pVertices^ := InitFont3DVertex(D3DXVector3(x+w,y+0,0), D3DXVector3(0,0,-1), tx2, ty2); Inc(pVertices);
      pVertices^ := InitFont3DVertex(D3DXVector3(x+0,y+h,0), D3DXVector3(0,0,-1), tx1, ty1); Inc(pVertices);
      dwNumTriangles := dwNumTriangles + 2;

      if (dwNumTriangles*3 > (MAX_NUM_VERTICES - 6)) then
      begin
        // Unlock, render, and relock the vertex buffer
        m_pVB.Unlock;
        m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLELIST, 0, dwNumTriangles);
        m_pVB.Lock(0, 0, PBYTE(pVertices), D3DLOCK_DISCARD);
        dwNumTriangles := 0;
      end;
    end;

    x := x + w;
    c:= strText^;
  end;

  // Unlock and render the vertex buffer
  m_pVB.Unlock;
  if (dwNumTriangles > 0) then
    m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLELIST, 0, dwNumTriangles);

  // Restore the modified renderstates
  m_pd3dDevice.ApplyStateBlock(m_dwSavedStateBlock);

  Result:= S_OK;
end;

procedure TD3DNitroFont.Draw(const X,Y: single; const Txt: String; Lev: Single;
  Flags: TFontRenderFlags;
  DoNotManageStates: Boolean);
Procedure DrawQuadRT(X, Y, Wid, Hgt, Lev, Tu, Tu2, Tv,Tv2: single);
var
  Quad: array[0..3] of TFont3DVertex;
begin
{  Tv := 1-Tv;
  Tv2 := 1-Tv2; }

  Quad[0].p:= D3DXVector3(X,     Y, -lev);
  Quad[1].p:= D3DXVector3(X+Wid, Y, -lev);
  Quad[2].p:= D3DXVector3(X+Wid, Y-Hgt, -lev);
  Quad[3].p:= D3DXVector3(X,     Y-Hgt, -lev);

  Quad[0].n:= D3DXVector3(0, 0, 1);
  Quad[1].n:= D3DXVector3(0, 0, 1);
  Quad[2].n:= D3DXVector3(0, 0, 1);
  Quad[3].n:= D3DXVector3(0, 0, 1);

  Quad[0].t1:= D3DXVector2(Tu,  Tv);
  Quad[1].t1:= D3DXVector2(Tu2, Tv);
  Quad[2].t1:= D3DXVector2(Tu2,Tv2);
  Quad[3].t1:= D3DXVector2(Tu, Tv2);

  m_pd3dDevice.DrawPrimitiveUP(D3DPT_TRIANGLEFAN, 2, Quad, SizeOf(TFont3DVertex))
end;

Var I, Let: integer;
    Xx: Single;
begin
  if txt = '' then exit;

  if not DoNotManageStates then
    if FAILED(PrepareDrawing(Flags)) then Exit;

  // Setup renderstate
  m_pd3dDevice.SetVertexShader(D3DFVF_Font3DVertex);

  // Turn off culling for two-sided text
  if (rfTwoSided in Flags) then
    m_pd3dDevice.SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);

  Xx := X;

  For I := 1 to length(Txt) do
  begin
    Let := Ord(Txt[I]) - StartChar;
    if (Let <> -1) and not (let > FontLen) then
    begin
      DrawQuadRT(
        Xx, Y, FontChars^[Let].W + 1.25, FontChars^[Let].H + 1.25, Lev,
        FontChars^[Let].T1, FontChars^[Let].T2 + 0.005,
        FontChars^[Let].U1, FontChars^[Let].U2 + 0.005);
      Xx := Xx + FontChars^[Let].W + 1.25 - CharOffset;
    end
    else
      Xx := Xx + SpaceWidth * fWidth{of Texture in pixels};
  end;

  // Restore the modified renderstates
  if not DoNotManageStates then
    m_pd3dDevice.ApplyStateBlock(m_dwSavedStateBlock);
end;

end.